/*****************************************************
Chip type               : ATmega8L
AVR Core Clock frequency: 8,000000 MHz
*****************************************************/

 //By controlling ship's throttle and orientation, you can take off, land, maneuver into circular and elliptical orbits. 


#include <mega8.h>
#include <alcd.h>
#include <math.h>
#include <STDLIB.H>
#include <DELAY.H>


#define CPU_freq 8000000
#define time_step 4

//planet constants
int G = 162;                           //free fall speedup on surface level  (1,62 m/(s*s) on the Moon)
long R = 1738000;                            //planet radius (1738000 m for Moon)
int C =  3660;                                   //outflow velocity of the products of combustion (3660 m/s for oxygen&kerosine based fuel)
int F_max = 3500;                                    //lander max onboard fuel mass (kg)  (3500kg max)
int M_wf  = 2250;                                    //lander mass without fuel (2250kg)
//unsigned long K;
long TMP;

bit int_t1, int_t1s;
unsigned char timer_1, timer_2;
char landing_quality;

struct craft
{
long V_y;                             //vehicle vertical velocity (m/s)
long V_x;                            //vehicle horisontal velocity (m/s)
long r;                               //vehicle vertical coordinate (m)
long y;                              //vehicle vertical coordinate to surfaces (m)
long x;                             //vehicle horisontal coordinate (m)
unsigned int throttle;
unsigned int orientation;
long fuel_mass;
long delta_fuel;
unsigned int orbital_velocity;
int speedup;
float angle;
}  lunolet = {0, 0, 1738000, 0, 0, 0, 0, 3500, 0, 0, 0, 0};

struct time
{
 char quarter_seconds;
 char seconds;
 char minutes;
 char hours;
} systime;

void update_time (struct time *t);
void display_time (struct time *t);

void update_craft_param (struct craft *j);
void display_craft_param (struct craft *j);

void ports_init ()
{
PORTB=0x00;
DDRB=0x00;

PORTC=0x00;
DDRC=0x00;

DDRD=0b00000000;
PORTD=0x00;  
}

void mask_prm_init ()
{
  lcd_gotoxy (0,0); lcd_putsf ("00000.0H00:00:00");
  lcd_gotoxy (0,1); lcd_putsf ("+000000.0S 00.0A");
  lcd_gotoxy (0,2); lcd_putsf ("0000+00000.0V00C");
  lcd_gotoxy (0,3); lcd_putsf ("0000-00000.0U000");
}

void ADC_init ()
{
 ADMUX |= 1<<REFS0 | 1<<REFS1 ;
 ADCSRA |= 1<<ADEN | 1<<ADPS0 | 1<<ADPS1 | 1<<ADPS2; // /128
}
void tmr_1_init ()
{
OCR1A = CPU_freq/4096;    // each 0,25s cm int
TCCR1B |= 1<<WGM12 | 1<<CS12 | 1<<CS10;  // CTC, OCR1A top, /1024
TIMSK |= 1<<OCIE1A;
}

interrupt [TIM1_COMPA] void tmr_1_cmpa (void)
{
update_time (&systime);
int_t1 = 1;

if (timer_1) timer_1--;
if (timer_2) timer_2--;
}



void main(void)
{
ports_init ();
tmr_1_init ();

ADC_init ();
lcd_init (16);
lcd_gotoxy (6,0);
lcd_putsf ("MOON");
lcd_gotoxy (5,1);
lcd_putsf ("LANDER");
lcd_gotoxy (6,2);
lcd_putsf ("GAME");
lcd_gotoxy (0,3);
lcd_putsf (">>>>>******<<<<<");
delay_ms (2000);
mask_prm_init ();





#asm ("sei")

while (1)
      {
       if (int_t1s) 
       {
       int_t1s = ~int_t1s; 
       display_time (&systime);

      
       }
       if (int_t1)
       {//ADC0_orientation, ADC1_throttle 
       if (ADMUX.MUX0) {ADMUX.MUX0 = 0; lunolet.throttle = ((long)ADCW*99)/1023; ADCSRA |=  1<<ADSC;}
       else {ADMUX.MUX0 = 1; lunolet.orientation = ((long)360*ADCW)/1023; ADCSRA |=  1<<ADSC;}
       update_craft_param (&lunolet);
       display_craft_param (&lunolet);
       int_t1 = 0;
 
       
       }
      }
}


void update_time (struct time *t)
{
t->quarter_seconds++;
if (t->quarter_seconds == 4) {t->quarter_seconds = 0; t->seconds ++; int_t1s = 1;}
if (t->seconds == 60) {t->seconds = 0; t->minutes ++;}
if (t->minutes == 60) {t->minutes = 0; t->hours ++;}
if (t->hours == 24) {t->hours = 0;}
}
void display_time (struct time *t)   
{
 lcd_gotoxy (8,0); lcd_putchar (t->hours/10 +0x30); lcd_putchar (t->hours%10 +0x30);
 lcd_gotoxy (11,0); lcd_putchar (t->minutes/10 +0x30); lcd_putchar (t->minutes%10 +0x30);
 lcd_gotoxy (14,0); lcd_putchar (t->seconds/10 +0x30); lcd_putchar (t->seconds%10 +0x30);
}

void update_craft_param (struct craft *j)
{
enum about_fuel {full, presence, empty} fuel_state;
enum about_flight {stay_on_ground, takeoff, fly, landing} flight_state;
bit landing;
long r_tmp; 

/*
long V_y;                             //vehicle vertical velocity (m/s)
long V_x;                            //vehicle horisontal velocity (m/s)
long r;                               //vehicle vertical coordinate (m)
long y;                              //vehicle vertical coordinate to surfaces (m)
long x;                             //vehicle horisontal coordinate (m)
unsigned int throttle;
unsigned int orientation;
long fuel_mass;
unsigned int orbital_velocity;
int speedup;
*/

    if ((j->y == 0)&&((labs (j->x)) < 50)) 
    {
    if (j->fuel_mass < (F_max - 10)) j->fuel_mass += 10;
    if (j->fuel_mass < F_max) j->fuel_mass += 1;
    if (j->fuel_mass == F_max) fuel_state = full;
    }
    else
    {
    if (j->fuel_mass >= (j->throttle/time_step)) 
    {
    j->fuel_mass = (float)(j->fuel_mass - ((float)j->throttle/time_step));
    fuel_state = presence;
    }
    else if (j->fuel_mass) {j->throttle = j->fuel_mass/time_step; j->fuel_mass = 0; fuel_state = empty;}
    else {fuel_state = empty; j->throttle = 0;}
    }
    
    j->speedup = ((float) j->throttle * C*100) /(M_wf + j->fuel_mass);
    if ((j->speedup > 9000)) {timer_1 = (j->speedup-9000)*4/100;} 
    if (timer_1) {j->throttle = 0;}
    
    j->angle = ((float)j->orientation/180)*PI; 
    
    //vertical speed and coordinate
    if (j->r > R) 
    { 
    TMP = (float)(j->V_y + ((float)((float) j->speedup * cos ((float) j->angle) - (((float)G*R*R)/((float)j->r*j->r)) +((float)j->V_x*j->V_x/j->r) )/((float)10*time_step)));
     j->r = j->r + ((j->V_y + TMP)/(time_step*2)); 
     
     if ((j->r - R)<= 0)
     {
      landing_quality = sqrt (j->V_x*j->V_x + j->V_y*j->V_y);
      if (landing_quality < 25) {lcd_gotoxy (0,0); lcd_putsf (">soft landing < ");}
      else if (landing_quality < 50) {lcd_gotoxy (0,0); lcd_putsf (">normal landing<");}
      else if (landing_quality < 125) {lcd_gotoxy (0,0); lcd_putsf (">hard landing<  ");}
      else  {lcd_gotoxy (0,0); lcd_putsf (">crash in Moon )");}
      delay_ms (8000);
      j->V_y = 0; j->V_x = 0; j->r = R;
      mask_prm_init (); 
 
     }
       
        lcd_gotoxy (10,1);
        if (j->V_y < TMP) {lcd_putchar ('+');}
        else if (j->V_y == TMP) {lcd_putchar (' ');}
        else  {lcd_putchar ('-');}
     
     j->V_y = TMP;
    }
    else if (j->r == R)
    {
        if ( (j->speedup * cos ((float) j->angle)) >= G )
        {
       TMP = (float)j->V_y + ((float)((float) j->speedup * cos ((float) j->angle) - G)/((float)10*time_step));
        j->r = j->r + (j->V_y + TMP)/(time_step*2);
        j->V_y = TMP;
        }
        else
        {
        TMP = 0; j->V_y = 0;
        }
       
   
    }
    else 
    {
     j->r = R; j->y = 0; j->V_y = 0;
     lcd_gotoxy (10,1); lcd_putchar (' ');
    }
     
     j->y = j->r - R;
    
    //horisontal speed and coordinate
    TMP = (float)j->V_x + ((float)(((float)j->speedup*sin((float)j->angle)))/((float)10*time_step));
    if (j->y) { j->V_x = TMP;}
    else {j->V_x = 0; TMP = 0;} 
    j->x = j->x + (float)(j->V_x + TMP)/(2*time_step);
   

    
(*j).orbital_velocity =  sqrt((float)((G*(float)R*R)/((*j).r*100)));

    
 



}
void display_craft_param (struct craft *j)
{
    
       lcd_gotoxy (4,2);
       
       if ((j->V_y > 0)&&((((labs (j->V_y))/100000)%10))) {lcd_putchar ('+');}
       else if (j->V_y == 0) {lcd_putchar (' ');}
       else if ((j->V_y < 0)&&(((labs (j->V_y))/100000)%10)) {lcd_putchar ('-');}
       else {lcd_putchar (' ');}
       
       if (((labs (j->V_y))/100000)%10) 
       { 
       lcd_putchar (((labs (j->V_y))/100000)%10 +0x30);
       lcd_putchar (((labs (j->V_y))/10000)%10 +0x30);
       lcd_putchar (((labs (j->V_y))/1000)%10 +0x30);
       lcd_putchar (((labs (j->V_y))/100)%10 +0x30);
       lcd_putchar (((labs (j->V_y))/10)%10 +0x30);
       lcd_putsf (".");
       lcd_putchar ((labs (j->V_y))%10 +0x30);
       }
       else 
       {
       if ((j->V_y < 0)&&(((labs (j->V_y))/10000)%10)) {lcd_putchar ('-');} 
       else if ((j->V_y > 0)&&((((labs (j->V_y))/10000)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
       if (((labs (j->V_y))/10000)%10)      
            {
            lcd_putchar (((labs (j->V_y))/10000)%10 +0x30);
            lcd_putchar (((labs (j->V_y))/1000)%10 +0x30);
            lcd_putchar (((labs (j->V_y))/100)%10 +0x30);
            lcd_putchar (((labs (j->V_y))/10)%10 +0x30);
            lcd_putsf (".");
            lcd_putchar ((labs (j->V_y))%10 +0x30);
            }
            else 
            {
                if ((j->V_y < 0)&&(((labs (j->V_y))/1000)%10)) {lcd_putchar ('-');}  
                else if ((j->V_y > 0)&&((((labs (j->V_y))/1000)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
                if (((labs (j->V_y))/1000)%10)      
                {
                lcd_putchar (((labs (j->V_y))/1000)%10 +0x30);
                lcd_putchar (((labs (j->V_y))/100)%10 +0x30);
                lcd_putchar (((labs (j->V_y))/10)%10 +0x30);
                lcd_putsf (".");
                lcd_putchar ((labs (j->V_y))%10 +0x30);
                }
                    else 
                    {
                    if ((j->V_y < 0)&&(((labs (j->V_y))/100)%10)) {lcd_putchar ('-');} 
                    else if ((j->V_y > 0)&&((((labs (j->V_y))/100)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
                    if (((labs (j->V_y))/100)%10)
                        {
                        lcd_putchar (((labs (j->V_y))/100)%10 +0x30);
                        lcd_putchar (((labs (j->V_y))/10)%10 +0x30);
                        lcd_putsf (".");
                        lcd_putchar ((labs (j->V_y))%10 +0x30);
                        }
                        else
                        {
                        if (j->V_y < 0) {lcd_putchar ('-');} 
                        else if (j->V_y > 0) {lcd_putchar ('+');} else {lcd_putchar (' ');}
                        lcd_putchar (((labs (j->V_y))/10)%10 +0x30);
                        lcd_putsf (".");
                        lcd_putchar ((labs (j->V_y))%10 +0x30);
                        }
                    }
           }
       } 
       


       lcd_gotoxy (4,3);
       if ((j->V_x > 0)&&((((labs (j->V_x))/100000)%10))) {lcd_putchar ('+');}
       else if (j->V_x == 0) {lcd_putchar (' ');}
       else if ((j->V_x < 0)&&(((labs (j->V_x))/100000)%10)) {lcd_putchar ('-');}
       else {lcd_putchar (' ');}
       
       if (((labs (j->V_x))/100000)%10) 
       { 
       lcd_putchar (((labs (j->V_x))/100000)%10 +0x30);
       lcd_putchar (((labs (j->V_x))/10000)%10 +0x30);
       lcd_putchar (((labs (j->V_x))/1000)%10 +0x30);
       lcd_putchar (((labs (j->V_x))/100)%10 +0x30);
       lcd_putchar (((labs (j->V_x))/10)%10 +0x30);
       lcd_putsf (".");
       lcd_putchar ((labs (j->V_x))%10 +0x30);
       }
       else 
       {
       if ((j->V_x < 0)&&(((labs (j->V_x))/10000)%10)) {lcd_putchar ('-');} 
       else if ((j->V_x > 0)&&((((labs (j->V_x))/10000)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
       if (((labs (j->V_x))/10000)%10)      
            {
            lcd_putchar (((labs (j->V_x))/10000)%10 +0x30);
            lcd_putchar (((labs (j->V_x))/1000)%10 +0x30);
            lcd_putchar (((labs (j->V_x))/100)%10 +0x30);
            lcd_putchar (((labs (j->V_x))/10)%10 +0x30);
            lcd_putsf (".");
            lcd_putchar ((labs (j->V_x))%10 +0x30);
            }
            else 
            {
                if ((j->V_x < 0)&&(((labs (j->V_x))/1000)%10)) {lcd_putchar ('-');}  
                else if ((j->V_x > 0)&&((((labs (j->V_x))/1000)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
                if (((labs (j->V_x))/1000)%10)      
                {
                lcd_putchar (((labs (j->V_x))/1000)%10 +0x30);
                lcd_putchar (((labs (j->V_x))/100)%10 +0x30);
                lcd_putchar (((labs (j->V_x))/10)%10 +0x30);
                lcd_putsf (".");
                lcd_putchar ((labs (j->V_x))%10 +0x30);
                }
                    else 
                    {
                    if ((j->V_x < 0)&&(((labs (j->V_x))/100)%10)) {lcd_putchar ('-');} 
                    else if ((j->V_x > 0)&&((((labs (j->V_x))/100)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
                    if (((labs (j->V_x))/100)%10)
                        {
                        lcd_putchar (((labs (j->V_x))/100)%10 +0x30);
                        lcd_putchar (((labs (j->V_x))/10)%10 +0x30);
                        lcd_putsf (".");
                        lcd_putchar ((labs (j->V_x))%10 +0x30);
                        }
                        else
                        {
                        if (j->V_x < 0) {lcd_putchar ('-');} 
                        else if (j->V_x > 0) {lcd_putchar ('+');} else {lcd_putchar (' ');}
                        lcd_putchar (((labs (j->V_x))/10)%10 +0x30);
                        lcd_putsf (".");
                        lcd_putchar ((labs (j->V_x))%10 +0x30);
                        }
                    }
           }
       }
      
       
       lcd_gotoxy (0,1); //horisontal coordinate 
       
       if ((j->x > 0)&&((((labs (j->x))/1000000)%10))) {lcd_putchar ('+');}
       else if (j->x == 0) {lcd_putchar (' ');}
       else if ((j->x < 0)&&(((labs (j->x))/1000000)%10)) {lcd_putchar ('-');}
       else {lcd_putchar (' ');}
       
        
       if (((labs (j->x))/1000000)%10) 
       { 
       lcd_putchar (((labs (j->x))/1000000)%10 +0x30);
       lcd_putchar (((labs (j->x))/100000)%10 +0x30);
       lcd_putchar (((labs (j->x))/10000)%10 +0x30);
       lcd_putchar (((labs (j->x))/1000)%10 +0x30);
       lcd_putchar (((labs (j->x))/100)%10 +0x30);
       lcd_putchar (((labs (j->x))/10)%10 +0x30);
       lcd_putsf (".");
       lcd_putchar ((labs (j->x))%10 +0x30);
       } 
       else 
       {
       if ((j->x < 0)&&(((labs (j->x))/100000)%10)) {lcd_putchar ('-');} 
       else if ((j->x > 0)&&((((labs (j->x))/100000)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
       if (((labs (j->x))/100000)%10) 
       { 
       lcd_putchar (((labs (j->x))/100000)%10 +0x30);
       lcd_putchar (((labs (j->x))/10000)%10 +0x30);
       lcd_putchar (((labs (j->x))/1000)%10 +0x30);
       lcd_putchar (((labs (j->x))/100)%10 +0x30);
       lcd_putchar (((labs (j->x))/10)%10 +0x30);
       lcd_putsf (".");
       lcd_putchar ((labs (j->x))%10 +0x30);
       }
       else 
       {
       if ((j->x < 0)&&(((labs (j->x))/10000)%10)) {lcd_putchar ('-');} 
       else if ((j->x > 0)&&((((labs (j->x))/10000)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
       if (((labs (j->x))/10000)%10)      
            {
            lcd_putchar (((labs (j->x))/10000)%10 +0x30);
            lcd_putchar (((labs (j->x))/1000)%10 +0x30);
            lcd_putchar (((labs (j->x))/100)%10 +0x30);
            lcd_putchar (((labs (j->x))/10)%10 +0x30);
            lcd_putsf (".");
            lcd_putchar ((labs (j->x))%10 +0x30);
            }
            else 
            {
                if ((j->x < 0)&&(((labs (j->x))/1000)%10)) {lcd_putchar ('-');}  
                else if ((j->x > 0)&&((((labs (j->x))/1000)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
                if (((labs (j->x))/1000)%10)      
                {
                lcd_putchar (((labs (j->x))/1000)%10 +0x30);
                lcd_putchar (((labs (j->x))/100)%10 +0x30);
                lcd_putchar (((labs (j->x))/10)%10 +0x30);
                lcd_putsf (".");
                lcd_putchar ((labs (j->x))%10 +0x30);
                }
                    else 
                    {
                    if ((j->x < 0)&&(((labs (j->x))/100)%10)) {lcd_putchar ('-');} 
                    else if ((j->x > 0)&&((((labs (j->x))/100)%10))) {lcd_putchar ('+');} else {lcd_putchar (' ');}
                    if (((labs (j->x))/100)%10)
                        {
                        lcd_putchar (((labs (j->x))/100)%10 +0x30);
                        lcd_putchar (((labs (j->x))/10)%10 +0x30);
                        lcd_putsf (".");
                        lcd_putchar ((labs (j->x))%10 +0x30);
                        }
                        else
                        {
                        if (j->x < 0) {lcd_putchar ('-');} 
                        else if (j->x > 0) {lcd_putchar ('+');} else {lcd_putchar (' ');}
                        lcd_putchar (((labs (j->x))/10)%10 +0x30);
                        lcd_putsf (".");
                        lcd_putchar ((labs (j->x))%10 +0x30);
                        }
                    }
           }
       }}
       
       lcd_gotoxy (0,0); //vertical coordinate
       if (((j->y/100000)%10) == 0) 
       {
            lcd_putchar (' ');
            if (((j->y/10000)%10) == 0) 
            {
            lcd_putchar (' ');
                if (((j->y/1000)%10) == 0)
                {
                    lcd_putchar (' ');
                    if (((j->y/100)%10) == 0)
                    {
                    lcd_putchar (' ');
                    lcd_putchar ((j->y/10)%10 +0x30);
                    lcd_putsf (".");
                    lcd_putchar (j->y%10 +0x30);
                    }
                    else 
                    {
                    lcd_putchar ((j->y/100)%10 +0x30);
                    lcd_putchar ((j->y/10)%10 +0x30);
                    lcd_putsf (".");
                    lcd_putchar (j->y%10 +0x30);
                    }
                }
                else 
                {
                lcd_putchar ((j->y/1000)%10 +0x30);
                lcd_putchar ((j->y/100)%10 +0x30);
                lcd_putchar ((j->y/10)%10 +0x30);
                lcd_putsf (".");
                lcd_putchar (j->y%10 +0x30);
                }
            

            }
            else 
            {
            lcd_putchar ((j->y/10000)%10 +0x30);
            lcd_putchar ((j->y/1000)%10 +0x30);
            lcd_putchar ((j->y/100)%10 +0x30);
            lcd_putchar ((j->y/10)%10 +0x30);
            lcd_putsf (".");
            lcd_putchar (j->y%10 +0x30);
            }
            
       }
       else 
       {
       lcd_putchar ((j->y/100000)%10 +0x30);
       lcd_putchar ((j->y/10000)%10 +0x30);
       lcd_putchar ((j->y/1000)%10 +0x30);
       lcd_putchar ((j->y/100)%10 +0x30);
       lcd_putchar ((j->y/10)%10 +0x30);
       lcd_putsf (".");
       lcd_putchar (j->y%10 +0x30);
       }
       
       lcd_gotoxy (13,2);
       if (timer_1)
       {
       lcd_putchar ('P');
       lcd_putchar ('L');
       }
       else 
       {
       lcd_putchar ((j->throttle/10)%10 +0x30);
       lcd_putchar (j->throttle%10 +0x30);
       } 
       
       lcd_gotoxy (13,3);
       lcd_putchar ((j->orientation/100)%10 +0x30);
       lcd_putchar ((j->orientation/10)%10 +0x30);
       lcd_putchar (j->orientation%10 +0x30);

       lcd_gotoxy (0,2); // orbital velocity at this orbit 
       lcd_putchar ((j->orbital_velocity/1000)%10 +0x30);
       lcd_putchar ((j->orbital_velocity/100)%10 +0x30);
       lcd_putchar ((j->orbital_velocity/10)%10 +0x30);
       lcd_putchar (j->orbital_velocity%10 +0x30);       
       
       lcd_gotoxy (0,3); //amount of fuel
       if ((j->fuel_mass/1000)%10)
       {
       lcd_putchar ((j->fuel_mass/1000)%10 +0x30);
       lcd_putchar ((j->fuel_mass/100)%10 +0x30);
       lcd_putchar ((j->fuel_mass/10)%10 +0x30);
       lcd_putchar (j->fuel_mass%10 +0x30);
       }
       else
       {
        lcd_putchar (' ');
       if ((j->fuel_mass/100)%10)
       {
       lcd_putchar ((j->fuel_mass/100)%10 +0x30);
       lcd_putchar ((j->fuel_mass/10)%10 +0x30);
       lcd_putchar (j->fuel_mass%10 +0x30);        
       }
       else
       {
        lcd_putchar (' '); 
        if ((j->fuel_mass/10)%10)
       {
       lcd_putchar ((j->fuel_mass/10)%10 +0x30);
       lcd_putchar (j->fuel_mass%10 +0x30);        
       }
        else
       {
        lcd_putchar (' '); 
        if (j->fuel_mass%10)
       {
       lcd_putchar (j->fuel_mass%10 +0x30);        
       }
       } } }
       
       lcd_gotoxy (11,1);

       if (labs (j->speedup) > 999) 
       {
       lcd_putchar ((labs (j->speedup)/1000)%10 +0x30);
       lcd_putchar ((labs (j->speedup)/100)%10 +0x30);
       lcd_putsf (".");
       lcd_putchar ((labs (j->speedup)/10)%10 +0x30);
       }
       else if ((labs (j->speedup)) > 99)
       {
       lcd_putchar ((labs (j->speedup)/100)%10 +0x30);
       lcd_putsf (".");
       lcd_putchar ((labs (j->speedup)/10)%10 +0x30);
       lcd_putchar (labs (j->speedup)%10 +0x30);                            
       }
       else
       {
       lcd_putsf ("0");
       lcd_putsf (".");
       lcd_putchar ((labs (j->speedup)/10)%10 +0x30);
       lcd_putchar (labs (j->speedup)%10 +0x30);
       }

}